# -*- coding: utf-8 -*-
# Copyright (C) 2014 Oliver Ainsworth

import sys


class Platform(object):
    """A Source server platform identifier

    This class provides utilities for representing Source server platforms
    as returned from a A2S_INFO request. Each platform is ultimately
    represented by one of the following integers:

    +-----+----------+
    | ID  | Platform |
    +=====+==========+
    | 108 | Linux    |
    +-----+----------+
    | 109 | Mac OS X |
    +-----+----------+
    | 111 | Mac OS X |
    +-----+----------+
    | 119 | Windows  |
    +-----+----------+
    """

    def __init__(self, value):
        """Initialise the platform identifier

        The given ``value`` will be mapped to a numeric identifier. If the
        value is already an integer it must then it must exist in the table
        above else ValueError is returned.

        If ``value`` is a one character long string then it's ordinal value
        as given by ``ord()`` is used. Alternately the string can be either
        of the following:

        * Linux
        * Mac OS X
        * Windows
        """
        if isinstance(value, str):
            if len(value) == 1:
                value = ord(value)
            else:
                # fmt: off
                value = {
                    'linux': 108,
                    'mac os x': 111,
                    'windows': 119,
                }.get(value.lower())
                # fmt: on
                if value is None:
                    raise ValueError("Couldn't convert string {!r} to valid " 'platform identifier'.format(value))
        if value not in {108, 109, 111, 119}:
            raise ValueError('Invalid platform identifier {!r}'.format(value))
        self.value = value

    def __repr__(self):
        return '<{self.__class__.__name__} ' "{self.value} '{self}'>".format(self=self)

    def __unicode__(self):
        # fmt: off
        return {
            108: 'Linux',
            109: 'Mac OS X',
            111: 'Mac OS X',
            119: 'Windows',
        }[self.value]
        # fmt: on

    def __str__(self):
        return self.__unicode__()

    def __bytes__(self):
        return self.__unicode__().encode(sys.getdefaultencoding())

    def __int__(self):
        return self.value

    def __eq__(self, other):
        """Check for equality between two platforms

        If ``other`` is not a Platform instance then an attempt is made to
        convert it to one using same approach as :meth:`__init__`. This means
        platforms can be compared against integers and strings. For example:

        .. code:: pycon

            >>>Platform(108) == "linux"
            True
            >>>Platform(109) == 109
            True
            >>>Platform(119) == "w"
            True

        Despite the fact there are two numerical identifers for Mac (109 and
        111) comparing either of them together will yield ``True``.

        .. code:: pycon

            >>>Platform(109) == Platform(111)
            True
        """
        if not isinstance(other, Platform):
            other = Platform(other)
        if self.value == 109 or self.value == 111:
            return other.value == 109 or other.value == 111
        else:
            return self.value == other.value

    @property
    def os_name(self):
        """Convenience mapping to names returned by ``os.name``"""
        # fmt: off
        return {
            108: 'posix',
            109: 'posix',
            111: 'posix',
            119: 'nt',
        }[self.value]
        # fmt: on


Platform.LINUX = Platform(108)
Platform.MAC_OS_X = Platform(111)
Platform.WINDOWS = Platform(119)


class ServerType(object):
    """A Source server platform identifier

    This class provides utilities for representing Source server types
    as returned from a A2S_INFO request. Each server type is ultimately
    represented by one of the following integers:

    +-----+---------------+
    | ID  | Server type   |
    +=====+===============+
    | 100 | Dedicated     |
    +-----+---------------+
    | 108 | Non-dedicated |
    +-----+---------------+
    | 112 | SourceTV      |
    +-----+---------------+
    """

    def __init__(self, value):
        """Initialise the server type identifier

        The given ``value`` will be mapped to a numeric identifier. If the
        value is already an integer it must then it must exist in the table
        above else ValueError is returned.

        If ``value`` is a one character long string then it's ordinal value
        as given by ``ord()`` is used. Alternately the string can be either
        of the following:

        * Dedicated
        * Non-Dedicated
        * SourceTV
        """
        if isinstance(value, str):
            if len(value) == 1:
                value = ord(value)
            else:
                # fmt: off
                value = {
                    'dedicated': 100,
                    'non-dedicated': 108,
                    'sourcetv': 112,
                }.get(value.lower())
                # fmt: on
                if value is None:
                    raise ValueError("Couldn't convert string {!r} to valid " 'server type identifier'.format(value))
        if value not in {100, 108, 112}:
            raise ValueError('Invalid server type identifier {!r}'.format(value))
        self.value = value

    def __repr__(self):
        return '<{self.__class__.__name__} ' "{self.value} '{self}'>".format(self=self)

    def __unicode__(self):
        # fmt: off
        return {
            100: 'Dedicated',
            108: 'Non-Dedicated',
            112: 'SourceTV',
        }[self.value]
        # fmt: on

    def __str__(self):
        return self.__unicode__()

    def __bytes__(self):
        return self.__unicode__().encode(sys.getdefaultencoding())

    def __int__(self):
        return self.value

    def __eq__(self, other):
        """Check for equality between two server types

        If ``other`` is not a ServerType instance then an attempt is made to
        convert it to one using same approach as :meth:`.__init__`. This means
        server types can be compared against integers and strings. For example:

        .. code:: pycon

            >>>Server(100) == "dedicated"
            True
            >>>Platform(108) == 108
            True
            >>>Platform(112) == "p"
            True
        """
        if not isinstance(other, ServerType):
            other = ServerType(other)
        return self.value == other.value

    @property
    def char(self):
        return chr(self.value)


ServerType.DEDICATED = ServerType(100)
ServerType.NON_DEDICATED = ServerType(108)
ServerType.SOURCETV = ServerType(112)



import asyncio
import logging
import os
from ipaddress import IPv4Address
from ipaddress import ip_address

import pyroute2

from . import config

logger = logging.getLogger(__name__)


def _get_addr_interface(addr: IPv4Address):
    ipdb = pyroute2.IPDB()
    for idx, addresses in ipdb.ipaddr.items():
        for ifaddr, _prefix in addresses:
            if ip_address(ifaddr) == addr:
                return ipdb.by_index[idx]['ifname']
    return None


def get_ebpf_program_run_args():  # noqa: C901
    args = []

    is_wide = False
    interface = None
    for _server_name, server in config.settings.servers:
        bind_ip = server.network.bind_ip

        if str(bind_ip) == '0.0.0.0':
            server_interface = None
            is_wide = True
        else:
            server_interface = _get_addr_interface(bind_ip)
            if server_interface is None:
                raise AssertionError(f"Can't get interface name for {bind_ip}")

        if interface is None:
            interface = server_interface

        if server_interface != interface:
            raise config.ConfigurationError(
                'Different interfaces dont supported yet: ' f'{server_interface} != {interface}'
            )

        server_port = server.network.server_port
        bind_port = server.network.bind_port

        if not server.network.ebpf_no_redirect:
            if is_wide:
                arg = f'{server_port}:{bind_port}'
            else:
                arg = f'{bind_ip}:{server_port}:{bind_port}'

            args += ['-p', arg]

    if is_wide:
        logger.warning("Wide interface is not supported yet. '0.0.0.0' will be interpreted like 'default interface'")

    if interface is not None:
        args += ['-i', interface]

    return args


async def run_ebpf_redirection():
    executable = config.ebpf.executable
    if not isinstance(executable, list):
        executable = [executable]

    cwd = None
    args = []

    if config.ebpf.script_path is not None:
        args.append(config.ebpf.script_path.name)
        cwd = config.ebpf.script_path.parent.as_posix()

    args += get_ebpf_program_run_args()

    logger.info('Run %s', executable + args)

    process = await asyncio.create_subprocess_exec(*executable, *args, stdout=asyncio.subprocess.PIPE, cwd=cwd)

    while True:
        data = await process.stdout.readline()
        if data:
            logger.info(data.decode().rstrip())

        if process.stdout.at_eof():
            break

    retcode = process.returncode
    if retcode != os.EX_OK:
        logger.exception('eBPF redirection exit with code %s', retcode)
        raise RuntimeError

    logger.info('eBPF redirection normally exit with 0 code')



import contextlib
import socket


def get_available_port():
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def is_port_available(port: int):
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        try:
            s.bind(('', port))
        except OverflowError:
            return False
        except OSError as exc:
            if 'Address already in use' in str(exc):
                return False
            raise

    return True



import asyncio
import logging
from contextlib import suppress

import uvloop
from pid.decorator import pidfile

from . import config
from .epbf import run_ebpf_redirection
from .proxy import QueryProxy

logger = logging.getLogger('sqproxy')


@pidfile('sqproxy', piddir=config.settings.piddir)
def run():
    uvloop.install()
    with suppress(KeyboardInterrupt, SystemExit):
        asyncio.run(_run_servers())


async def _run_servers():
    if not config.servers:
        logger.warning('No one server to run. Please check config')
        return

    proxies = []
    for name, server in config.servers:
        if server.entrypoint is None:
            entrypoint = QueryProxy
        else:
            entrypoint = server.entrypoint.obj

        proxy = entrypoint(server, name=name)
        proxies.append(proxy)

    futures = [asyncio.ensure_future(proxy.run()) for proxy in proxies]

    if config.ebpf and config.ebpf.enabled:
        logger.info('eBPF redirection enabled')
        logger.info('Wait all proxies to be ready ...')
        await asyncio.gather(
            *[proxy.wait_ready() for proxy in proxies],
        )
        logger.info('Wait all proxies to be ready ... Done!')
        futures.append(asyncio.ensure_future(run_ebpf_redirection()))
    else:
        logger.info('eBPF redirection disabled')

    await asyncio.gather(*futures)


if __name__ == '__main__':
    run()



import asyncio
import collections
import functools
import logging
import random
import time
import typing

import async_timeout
import backoff

from . import config
from .source import messages
from .transport import bind
from .transport import connect

MAX_SIZE_32 = 2 ** 31 - 1

NO_RESPONSE = object()


class AwaitableDict(collections.UserDict):
    """Оборачивает все значения в asyncio.Future
    таким образом можно дождаться появления нужного значения в словаре
    """

    def __setitem__(self, key, value):
        if key in self:
            fut = self.data[key]
            if fut.done():
                fut = self.data[key] = asyncio.Future()
            fut.set_result(value)
        else:
            fut = asyncio.Future()
            fut.set_result(value)
            self.data[key] = fut

    def __getitem__(self, key):
        value = super().__getitem__(key)
        if not value.done():
            raise KeyError(key)
        return value.result()

    async def get_wait(self, key):
        if key not in self.data:
            fut = asyncio.Future()
            self.data[key] = fut
        else:
            fut = self.data[key]

        return await fut


class QueryProxy:
    A2S_EMPTY_CHALLENGE = -1
    connect = functools.partial(connect)
    bind = functools.partial(bind)

    def __init__(self, settings: config.ServerModel, name: str = None):
        listen_addr = (str(settings.network.bind_ip), settings.network.bind_port)
        server_addr = (str(settings.network.server_ip), settings.network.server_port)

        if name is None:
            name = '%s:%s' % listen_addr

        self.listen_addr = listen_addr
        self.server_addr = server_addr
        self.resp_cache = {}
        self.our_a2s_challenge = random.randint(1, MAX_SIZE_32)
        self.settings = settings
        self.logger = logging.getLogger(name)

    # noinspection PyPep8Naming
    @property
    def retry_TimeoutError(self):  # noqa: ignore=N802
        return backoff.on_exception(backoff.constant, asyncio.TimeoutError, logger=self.logger)

    async def listen_client_requests(self):
        self.logger.info('Binding ... ')
        async with (await self.bind(self.listen_addr)) as listening:
            self.logger.info('Binding ... done!')
            self.logger.info('Listen for client requests ...')
            while True:
                request, data, addr = await listening.recv_packet()
                if request is None:
                    self.logger.warning(
                        'Broken data was received: data[:150]=%s',
                        data[:150],
                    )
                    continue

                if addr[1] == 0:
                    # FIXME: https://github.com/MagicStack/uvloop/issues/338
                    continue

                response = self.get_response_for(request, None)
                if response is None:
                    self.logger.warning('No response for %s', request)
                    continue
                if response is NO_RESPONSE:
                    continue

                await listening.send_packet(response, addr=addr)

    async def update_server_query_cache(self):
        coros = [
            self.retry_TimeoutError(self._update_info)(),
            self.retry_TimeoutError(self._update_players)(),
        ]
        if not self.settings.no_a2s_rules:
            coros.append(self.retry_TimeoutError(self._update_rules)())

        return await asyncio.gather(*coros)

    async def send_recv_packet(self, client, packet: messages.Packet, timeout=None):
        """Send packet and wait for response for it

        In addition to call client.[send_packet(), recv_packet()] this method handle
        GetChallengeResponse logic

        :param client: connected client
        :param packet: any `messages.Packet` instance to send to
        :param timeout: how much wait response
            trigger asyncio.TimeoutError on exceeded
        :return: tuple (message, data, addr, new_challenge)
            `new_challenge` will be None if not present
        """
        old_challenge = packet.get('challenge')

        a2s_challenge = old_challenge
        while True:
            if a2s_challenge is not None:
                await client.send_packet(packet.encode(challenge=a2s_challenge))
            else:
                await client.send_packet(packet.encode())

            start = time.monotonic()
            with async_timeout.timeout(timeout):
                message, data, addr = await client.recv_packet()
                self.logger.debug('Got %s for %ss', message.__class__.__name__, time.monotonic() - start)

            if isinstance(message, messages.GetChallengeResponse):
                if old_challenge not in (self.A2S_EMPTY_CHALLENGE, None):
                    self.logger.warning(
                        'Challenge number changed: %s -> %s',
                        old_challenge,
                        message['challenge'],
                    )

                old_challenge = a2s_challenge
                a2s_challenge = message['challenge']
                continue

            break

        return message, data, addr, a2s_challenge

    async def _update_info(self):
        logger = self.logger.getChild('update-info')
        request = messages.InfoRequestV2()

        while True:
            async with (await connect(self.server_addr)) as client:
                logger.debug('Connected to %s (client port=%s)', self.server_addr, client.sockname[1])

                message, data, addr, a2s_challenge = await self.send_recv_packet(
                    client,
                    request,
                    timeout=max(
                        self.settings.a2s_response_timeout,
                        self.settings.a2s_info_cache_lifetime,
                    ),
                )

                self.resp_cache['a2s_info'] = data
                await asyncio.sleep(self.settings.a2s_info_cache_lifetime)

    async def _update_rules(self):
        logger = self.logger.getChild('update-rules')

        while True:
            async with (await connect(self.server_addr)) as client:
                logger.debug('Connected to %s (client port=%s)', self.server_addr, client.sockname[1])

                request = messages.RulesRequest(challenge=self.A2S_EMPTY_CHALLENGE)

                message, data, addr, a2s_challenge = await self.send_recv_packet(
                    client,
                    request,
                    timeout=max(
                        self.settings.a2s_response_timeout,
                        self.settings.a2s_rules_cache_lifetime,
                    ),
                )

                self.resp_cache['a2s_rules'] = data
                await asyncio.sleep(self.settings.a2s_rules_cache_lifetime)

    async def _update_players(self):
        logger = self.logger.getChild('update-players')

        while True:
            async with (await connect(self.server_addr)) as client:
                logger.debug('Connected to %s (client port=%s)', self.server_addr, client.sockname[1])

                request = messages.PlayersRequest(challenge=self.A2S_EMPTY_CHALLENGE)
                message, data, addr, a2s_challenge = await self.send_recv_packet(
                    client,
                    request,
                    timeout=max(
                        self.settings.a2s_response_timeout,
                        self.settings.a2s_players_cache_lifetime,
                    ),
                )

                self.resp_cache['a2s_players'] = data
                await asyncio.sleep(self.settings.a2s_players_cache_lifetime)

    def get_response_for(self, message, default) -> typing.Optional[bytes]:
        resp = default

        if isinstance(message, messages.InfoRequest):
            resp = self.resp_cache.get('a2s_info')
        elif isinstance(message, (messages.PlayersRequest, messages.RulesRequest)):
            challenge = message['challenge']

            if challenge == self.our_a2s_challenge:
                if isinstance(message, messages.PlayersRequest):
                    resp = self.resp_cache.get('a2s_players')
                elif isinstance(message, messages.RulesRequest):
                    resp = self.resp_cache.get('a2s_rules')
            elif self.our_a2s_challenge != self.A2S_EMPTY_CHALLENGE:
                # player request challenge number or we don't know who is it
                # return challenge number
                resp = messages.GetChallengeResponse(challenge=self.our_a2s_challenge).encode()

        return resp

    async def run(self):
        tasks = [
            asyncio.create_task(coro) for coro in (self.update_server_query_cache(), self.listen_client_requests())
        ]
        done, pending = await asyncio.wait(
            tasks,
            return_when=asyncio.FIRST_COMPLETED,
        )
        for task in done:
            exc = task.exception() if not task.cancelled() else None
            self.logger.error('Task unexpectedly completed', exc_info=exc)  # noqa: ignore=G201

        for task in pending:
            task.cancel()

    async def wait_ready(self):
        """Wait until all internals being ready to start"""
        resp_cache = self.resp_cache = AwaitableDict(self.resp_cache)

        await resp_cache.get_wait('a2s_info')
        await resp_cache.get_wait('a2s_players')
        if not self.settings.no_a2s_rules:
            await resp_cache.get_wait('a2s_rules')

        self.resp_cache = dict(resp_cache)



import importlib.util
import logging
import pathlib
import typing
from ipaddress import IPv4Address

import sentry_sdk
import yaml
from cached_property import cached_property
from pydantic import AnyHttpUrl
from pydantic import BaseModel
from pydantic import BaseSettings
from pydantic import Extra
from pydantic import confloat
from pydantic import conint
from pydantic import validator
from sentry_sdk.integrations.logging import LoggingIntegration

from . import __version__
from . import utils
from .dict_merge import dict_merge
from .logging import setup_logging

if typing.TYPE_CHECKING:
    from .proxy import QueryProxy

    QueryProxyType = typing.TypeVar('QueryProxyType', bound=QueryProxy)
else:
    QueryProxyType = None


__all__ = (
    'NetworkModel',
    'ServerModel',
    'Settings',
    'ConfigurationError',
    'settings',
)

logger = logging.getLogger('sqproxy.config')


class ConfigurationError(Exception):
    pass


class NetworkModel(BaseModel):
    server_ip: IPv4Address
    server_port: conint(ge=1, le=65535)
    bind_ip: typing.Optional[IPv4Address] = None
    bind_port: typing.Optional[conint(ge=0, le=65535)] = 0
    ebpf_no_redirect: bool = False

    class Config:
        extra = Extra.forbid

    @validator('bind_ip', always=True)
    def _set_default_bind_ip(cls, bind_ip, values):
        if not bind_ip:
            bind_ip = values['server_ip']

        return bind_ip

    @validator('bind_port', always=True)
    def _set_default_bind_port(cls, bind_port, values):
        if bind_port is None or bind_port == 0:
            server_port = values['server_port']

            if utils.is_port_available(server_port + 800):
                # try to use pretty port
                bind_port = server_port + 800
            else:
                bind_port = utils.get_available_port()

        return bind_port


class EntrypointModel(BaseModel):
    """Allow override proxy server entrypoint"""

    path: str  #: path to callable which should return :class:`QueryProxy`-based object
    obj: typing.Optional[typing.Callable[..., QueryProxyType]] = None

    class Config:
        extra = Extra.forbid

    @validator('obj', always=True)
    def _import_obj(cls, v, values):
        if v is not None:
            return v

        path: str = values['path']
        if ':' not in path:
            raise ValueError(f'entry not found, expected path(s): example.py:MyObj, example:MyObj. Given: {path}')

        file_path, _, attr_name = path.rpartition(':')

        if not file_path.isidentifier():
            file_path = pathlib.Path(file_path).resolve()
            spec = importlib.util.spec_from_file_location('entrypoint', file_path)
            module = importlib.util.module_from_spec(spec)
            module.__loader__.exec_module(module)  # noqa
        else:
            module = importlib.import_module(file_path)

        try:
            return getattr(module, attr_name)
        except AttributeError as exc:
            raise AttributeError(exc.args[0].replace("'entrypoint'", f"'{file_path}'"))


class ServerModel(BaseModel):
    meta: dict
    network: NetworkModel
    a2s_info_cache_lifetime: confloat(gt=0) = 5
    a2s_rules_cache_lifetime: confloat(gt=0) = 5
    a2s_players_cache_lifetime: confloat(gt=0) = 1
    a2s_response_timeout: confloat(gt=0) = 1
    no_a2s_rules: bool = False
    entrypoint: typing.Optional[EntrypointModel] = None

    @validator('entrypoint', pre=True)
    def _entrypoint(cls, v, values):
        if isinstance(v, str):
            v = {'path': v}

        v['path'] = v['path'].format_map(values['meta'])
        return v

    class Config:
        extra = Extra.allow


class EBPFModel(BaseModel):
    enabled: bool = False
    executable: typing.Union[str, typing.List[str]] = 'python2'
    script_path: typing.Optional[pathlib.Path] = None

    class Config:
        extra = Extra.forbid


NamedServersType = typing.List[typing.Tuple[str, ServerModel]]


class Settings(BaseSettings):
    sentry_dsn: typing.Optional[AnyHttpUrl] = None
    confdir_0: pathlib.Path = '/etc/sqproxy/conf.d/'
    confdir_1: pathlib.Path = './conf.d/'
    error_log: pathlib.Path = '/dev/null'
    loglevel: str = 'INFO'
    piddir: typing.Optional[pathlib.Path] = None

    class Config:
        env_file = '.env'
        env_prefix = 'SQPROXY_'
        keep_untouched = (cached_property,)

    @cached_property
    def merged_config_data(self):
        return load_configs(iter_config_files(self.confdir_0, self.confdir_1))

    @validator('loglevel')
    def _check_loglevel(cls, v):
        from logging import _checkLevel  # noqa

        _checkLevel(v)
        return v

    @cached_property
    def servers(self) -> NamedServersType:
        merged_config_data = self.merged_config_data

        servers = merged_config_data.get('servers')
        if servers:
            servers = [(name, ServerModel.parse_obj(server)) for name, server in servers.items()]

        return servers or []

    @cached_property
    def ebpf(self) -> typing.Optional[EBPFModel]:
        ebpf = self.merged_config_data.get('ebpf')
        if not ebpf:
            return None
        return EBPFModel.parse_obj(ebpf)


def _apply_defaults(target, defaults):
    target.update(dict_merge(defaults, target))


def _get_old_config(data: dict, global_defaults) -> typing.Tuple[typing.Dict, typing.Dict]:
    """Get full and ready to use config"""
    defaults = data.pop('defaults', None) or {}

    servers_data = data.get('servers')
    if not servers_data:
        return data, defaults

    for server in servers_data.values():
        for g_defaults in global_defaults:
            _apply_defaults(server, g_defaults['values'])

        if defaults and defaults.get('values'):
            _apply_defaults(server, defaults['values'])

    return data, defaults


def _get_config(data: dict, global_defaults) -> typing.Tuple[typing.Dict, typing.Dict]:
    """Get full and ready to use config"""
    if 'global' in (data.get('defaults') or {}):
        data, defaults = _get_old_config(data, global_defaults)
        is_global = defaults.pop('global')
        defaults = defaults['values']
        defaults['__global__'] = is_global
        return data, defaults

    defaults = data.pop('defaults', None) or {}

    servers_data = data.get('servers')
    if not servers_data:
        return data, defaults

    for server in servers_data.values():
        for g_defaults in global_defaults:
            _apply_defaults(server, g_defaults)

        if defaults:
            _apply_defaults(server, defaults)

    return data, defaults


def iter_config_files(*confdirs):
    for confdir in confdirs:
        if not confdir.exists():
            logger.info('Confdir not found: %s', confdir.absolute().as_posix())
            continue

        if not confdir.is_dir():
            logger.warning('Confdir expected to be a directory, not a file: %s', confdir.absolute().as_posix())
            continue

        for file in sorted(confdir.iterdir(), key=lambda f: f.name):
            if file.is_file() and file.name.endswith('.yaml'):
                logger.info('Found config: %s', file.as_posix())
                yield file
            else:
                logger.debug('Found non-config file, ignore: %s', file.as_posix())


def load_configs(paths: typing.Iterable[pathlib.Path]):
    configs = []
    global_defaults = []

    ebpf_configured = False

    for path in paths:
        with path.open() as fp:
            config, config_defaults = _get_config(yaml.full_load(fp), global_defaults)

            if 'ebpf' in config:
                if ebpf_configured:
                    raise ConfigurationError('eBPF already configured')
                ebpf_configured = True

            if config_defaults.pop('__global__', False):
                global_defaults.append(config_defaults)

            if config:
                meta = {'CONFPATH': path, 'CONFDIR': path.parent}
                for server in (config.get('servers') or {}).values():
                    server['meta'] = meta

                configs.append(config)

    whole_config = {}
    for config in configs:
        whole_config = dict_merge(whole_config, config)

    return whole_config


def setup(settings_: Settings = None, reread: bool = False):
    global settings
    if settings_ is None or reread:
        logger.info('Re-read settings')
        settings_ = Settings()

    settings = settings_

    if settings.sentry_dsn:
        sentry_logging = LoggingIntegration(
            level=logging.DEBUG,  # Capture info and above as breadcrumbs
            event_level=logging.ERROR,  # Send errors as events
        )
        sentry_sdk.init(
            dsn=settings.sentry_dsn,
            integrations=[sentry_logging],
            release=__version__,
        )

    setup_logging(loglevel=settings.loglevel, error_filename=settings.error_log.as_posix())

    if settings.sentry_dsn:
        logger.debug('Sentry enabled')
    else:
        logger.debug('Sentry disabled')

    return settings


settings = Settings()

setup(settings)


def __getattr__(name):
    if name == 'ebpf':
        return settings.ebpf
    elif name == 'servers':
        return settings.servers
    else:
        raise AttributeError(name)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['source_query_proxy', 'source_query_proxy.source']

package_data = \
{'': ['*']}

install_requires = \
['async-timeout>=3.0,<4.0',
 'asyncio_dgram>=2.1.0,<3.0.0',
 'backoff>=1.10,<2.0',
 'cached-property>=1.5.2,<2.0.0',
 'click>=7.0,<8.0',
 'pid>=2.2,<4.0',
 'pydantic[dotenv]>=1.4,<2.0',
 'pylru>=1.2.0,<2.0.0',
 'pyroute2>=0.7.5,<0.8.0',
 'python-dotenv>=0.10.3,<0.15.0',
 'pyyaml>=6.0,<7.0',
 'sentry-sdk>=0.14.3,<0.20.0',
 'uvloop>=0.16.0,<0.17.0']

entry_points = \
{'console_scripts': ['sqproxy = source_query_proxy.cli:sqproxy']}

setup_kwargs = {
    'name': 'source-query-proxy',
    'version': '2.1.2',
    'description': 'Async proxy for Source Engine Query Protocol',
    'long_description': "\nsource-query-proxy\n==================\n\nMotivation\n----------\n\nBasically Source game-servers works in one thread and can't use more than one core for in-game logic.\nFor example - CS:GO, CS:Source, Left 4 Dead 2, etc.\n\nYes, you can use SourceMod to offload calculations (use threading), but we talking about common game logic.\nE.g. you can try use `DoS Protection extension <https://forums.alliedmods.net/showpost.php?p=2518787&postcount=117>`_, but caching is not fast solution, cause server spent time to receiving and sending answer from cache.\n\nThis solution allow redirect some (A2S query) packets to backend and game server don't spent time to answer anymore.\n\n\nIPTables (or any NAT) can't help!\n---------------------------------\n\nIf you use IPTables (NAT) to redirect queries to proxy, this rule will be remembered in routing table and if client try to connect - connection will be redirected to proxy too.\n\nWe use right way to redirect - eBPF: https://github.com/sqproxy/sqredirect\n\nPrerequisites\n-------------\n\nPython 3.7 or above\n\nYou can use `pyenv <https://github.com/pyenv/pyenv>`_ to install any version of Python without root privileges\n\nInstalling\n----------\n\n.. code-block:: bash\n\n    pip install source-query-proxy==2.1.2\n\nConfiguring\n-----------\n\nsqproxy search configs in ``/etc/sqproxy/conf.d`` and ``./conf.d`` directories.\nYou should place your config files only in this directories.\n\nFor more info see `examples <examples/conf.d>`_\n\nRun\n---\n\n.. code-block:: bash\n\n    sqproxy run\n\n\nRun with eBPF\n-------------\n\nPlease read the instruction and install: https://github.com/sqproxy/sqredirect\n\n1. Enable eBPF in config (see ``examples/00-globals.yaml``)\n\n2. Run\n\n.. code-block:: bash\n\n    sqproxy run\n\nRun daemonized via systemd\n--------------------------\n\n1. Copy the systemd unit file ``examples/systemd/system/sqproxy.service`` to ``/etc/systemd/system/sqproxy.service`` (Optional: Adjust the ``ExecStart`` path if you have installed sqproxy into a different directory)\n\n2. Enable and start the service\n\n.. code-block:: bash\n\n    systemctl enable --now sqproxy.service\n\n\nDevelopment\n-----------\n\n.. code-block:: bash\n\n    git clone https://github.com/spumer/source-query-proxy.git\n    cd source-query-proxy\n    poetry install\n    \n\nCredits\n-------\n\nSource Engine messages inspired by **Python-valve**\nhttps://github.com/serverstf/python-valve\n\n",
    'author': 'spumer',
    'author_email': 'spumer-tm@yandex.ru',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/sqproxy/sqproxy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# encoding: utf-8

import time


class IntervalCounter:
    def __init__(self, interval=1):
        self.interval = interval

        self._prev_count = 0
        self._current_count = 0

        self.marker = 0

    def _update(self, marker=None):
        if marker is None:
            marker = int(time.perf_counter())

        if marker - self.interval >= self.marker:
            self._prev_count = self._current_count
            self._current_count = 0
            self.marker = marker

    @property
    def count(self):
        self._update()
        return self._prev_count

    def emit(self):
        self._update()
        self._current_count += 1


class CSGOAggregator:
    def __init__(self):
        self.mps_counter = IntervalCounter()

    def aggregate(self, addr, msg=None):
        self.mps_counter.emit()

    def format_lines(self):
        lines = [
            'Messages per second: {mps} m/s'.format(mps=self.mps_counter.count),
        ]

        return lines

    def format_stats(self):
        return '\n'.join(self.format_lines())


def test():
    counter = IntervalCounter(interval=5)

    for _ in range(10000):
        counter.emit()
        time.sleep(0.1)

        print('Counter:', counter.count, end='\r')  # noqa: ignore=T001


if __name__ == '__main__':
    test()
